import pytest

import urllib

import torch
from PIL import Image
from torchvision import transforms

import torchtrail


@pytest.mark.parametrize("show_modules", [True, False])
def test_resnet(tmp_path, show_modules):
    model = torch.hub.load("pytorch/vision:v0.10.0", "resnet18", pretrained=True).eval()

    url = "https://github.com/pytorch/hub/raw/master/images/dog.jpg"
    filename = tmp_path / "dog.jpg"

    try:
        urllib.URLopener().retrieve(url, filename)
    except Exception:
        urllib.request.urlretrieve(url, filename)

    input_image = Image.open(filename)
    preprocess = transforms.Compose(
        [
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ]
    )
    input_tensor = preprocess(input_image).unsqueeze(0)

    with torchtrail.trace():
        input_tensor = torch.as_tensor(input_tensor)
        output_tensor = model(input_tensor)
        output_tensor = torch.nn.functional.softmax(output_tensor[0], dim=0)

    torchtrail.visualize(
        output_tensor, show_modules=show_modules, file_name=tmp_path / "resnet18.svg"
    )
    assert len(torchtrail.get_graph(output_tensor)) == 4
    assert len(torchtrail.get_graph(output_tensor, flatten=True)) == 174



import torch

import torchtrail


def test_rand_exp():
    with torchtrail.trace():
        input_tensor = torch.rand(1, 64)
        output_tensor = torch.exp(input_tensor)

    torchtrail.visualize(output_tensor)
    assert len(torchtrail.get_graph(output_tensor)) == 2


def test_rand_add_exp():
    with torchtrail.trace():
        input_tensor = torch.rand(1, 64)
        output_tensor = input_tensor + input_tensor
        output_tensor = torch.exp(output_tensor)

    torchtrail.visualize(output_tensor)
    assert len(torchtrail.get_graph(output_tensor)) == 3


def test_zeros():
    with torchtrail.trace():
        input_tensor = torch.zeros(1, 64)

    torchtrail.visualize(input_tensor)
    assert len(torchtrail.get_graph(input_tensor)) == 1


def test_rand_split():
    with torchtrail.trace():
        input_tensor = torch.rand(1, 64)
        output_tensors = torch.split(input_tensor, split_size_or_sections=32, dim=1)

    torchtrail.visualize(output_tensors)
    assert len(torchtrail.get_graph(output_tensors)) == 2


def test_rand_exp_backward():
    with torchtrail.trace():
        input_tensor = torch.rand(1, 64)
        input_tensor.requires_grad = True
        output_tensor = torch.exp(input_tensor)
        output_tensor.backward(torch.ones_like(output_tensor))

    torchtrail.visualize(input_tensor.grad)
    assert len(torchtrail.get_graph(input_tensor.grad)) == 2



import pytest

import torch
import transformers

import torchtrail


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit_embeddings(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    num_channels = 3
    height = 224
    width = 224

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTEmbeddings(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, num_channels, height, width)
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit_embeddings.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) == 11


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit_self_attention(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    sequence_size = 197
    hidden_size = 768

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTSelfAttention(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, sequence_size, hidden_size)
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit_self_attention.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) == 25


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit_self_output(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    sequence_size = 197
    hidden_size = 768

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTSelfOutput(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, sequence_size, hidden_size)
        residual_input_tensor = torch.randn(batch_size, sequence_size, hidden_size)
        output = model(input_tensor, residual_input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit_self_output.svg"
    )
    assert len(torchtrail.get_graph(output)) == 3
    assert len(torchtrail.get_graph(output, flatten=True)) in {6, 7}


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit_attention(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    sequence_size = 197
    hidden_size = 768

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTAttention(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, sequence_size, hidden_size)
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit_attention.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) in {29, 30}


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit_intermediate(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    sequence_size = 197
    hidden_size = 768

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTIntermediate(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, sequence_size, hidden_size)
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit_intermediate.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) == 5


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit_layer(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    sequence_size = 197
    hidden_size = 768

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTLayer(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, sequence_size, hidden_size)
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit_layer.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) in {45, 46}


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit_encoder(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    sequence_size = 197
    hidden_size = 768

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTEncoder(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, sequence_size, hidden_size)
        output = model(input_tensor).last_hidden_state

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit_encoder.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) in {
        541 - config.num_hidden_layers,
        541,
    }


@pytest.mark.parametrize("show_modules", [True, False])
def test_vit(tmp_path, show_modules):
    model_name = "google/vit-base-patch16-224"
    batch_size = 1
    num_channels = 3
    height = 224
    width = 224

    config = transformers.ViTConfig.from_pretrained(model_name)
    model = transformers.models.vit.modeling_vit.ViTModel(config).eval()

    with torchtrail.trace():
        input_tensor = torch.randn(batch_size, num_channels, height, width)
        output = model(input_tensor).pooler_output

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "vit.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) in {
        559 - config.num_hidden_layers,
        559,
    }



# MIT License

# Copyright (c) 2024 Akhmed Rakhmati

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from collections.abc import Iterable
import pathlib
from typing import Optional, Union

from loguru import logger
import graphviz
import networkx
from pyrsistent import PClass, field, pmap_field, PMap, pmap

from networkx.classes.coreviews import MultiAdjacencyView
from networkx.classes.reportviews import (
    DiMultiDegreeView,
    InMultiDegreeView,
    OutMultiDegreeView,
    NodeView,
    InMultiEdgeView,
    OutMultiEdgeView,
)


class MultiDiGraph(PClass):
    _node = pmap_field(object, PMap)
    _succ = pmap_field(object, PMap)
    _pred = pmap_field(object, PMap)
    _attributes = field(PMap, initial=pmap)

    @property
    def _adj(self):
        return self._succ

    @property
    def adj(self):
        return MultiAdjacencyView(self._adj)

    @property
    def succ(self):
        return self._succ

    @property
    def pred(self):
        return self._pred

    @property
    def graph(self):
        return self._attributes

    def __len__(self):
        return len(self._node)

    def __iter__(self):
        return iter(self._node)

    def __getitem__(self, node):
        return self.adj[node]

    def __contains__(self, node):
        return node in self._node

    def __hash__(self):
        return hash((self._node, self._pred, self._succ, self._attributes))

    def is_directed(self) -> bool:
        return True

    def is_multigraph(self) -> bool:
        return True

    def add_attributes(self, **kwargs) -> "MultiDiGraph":
        _attributes = self._attributes.update(kwargs)
        return self.set(_attributes=_attributes)

    def has_node(self, node):
        return node in self

    def add_node(self, node, **kwargs):
        attributes = self._node.get(node, pmap())
        attributes = attributes.update(pmap(kwargs))
        _node = self._node.set(node, attributes)

        _pred = self._pred
        if node not in _pred:
            _pred = _pred.set(node, pmap())

        _succ = self._succ
        if node not in _succ:
            _succ = _succ.set(node, pmap())

        new_graph = self.set(_node=_node, _pred=_pred, _succ=_succ)
        return new_graph

    def add_nodes_from(self, nodes_for_adding, **new_attributes):
        new_graph = self
        for node in nodes_for_adding:
            if isinstance(node, tuple):
                node, ndict = node
                newdict = pmap(new_attributes).update(ndict)
            else:
                newdict = pmap(new_attributes)
            attributes = self._node[node] if node in self._node else pmap()
            attributes = attributes.update(newdict)
            new_graph = new_graph.add_node(node, **attributes)
        return new_graph

    def add_edge(self, source, sink, key=None, **kwargs):
        if source not in self:
            self = self.add_node(source)
        if sink not in self:
            self = self.add_node(sink)

        _node = self._node

        def _add_edge(node_to_neighbors, from_node, to_node, edge_key):
            neighbors = node_to_neighbors.get(from_node, pmap())
            edges = neighbors.get(to_node, pmap())
            if edge_key is None:
                edge_key = max(edges.keys()) + 1 if edges.keys() else 0
            edges = edges.set(edge_key, pmap(kwargs))
            neighbors = neighbors.set(to_node, edges)
            node_to_neighbors = node_to_neighbors.set(from_node, neighbors)
            return node_to_neighbors

        _pred = _add_edge(self._pred, sink, source, key)
        _succ = _add_edge(self._succ, source, sink, key)

        new_graph = self.set(_node=_node, _pred=_pred, _succ=_succ)
        return new_graph

    def add_edges_from(self, ebunch_to_add, **attr):
        new_graph = self
        for edge in ebunch_to_add:
            if len(edge) == 4:
                source, sink, key, data = edge
            elif len(edge) == 3:
                source, sink, data = edge
                key = None
            elif len(edge) == 2:
                source, sink = edge
                key = None
                data = {}
            else:
                raise networkx.NetworkXError(
                    f"Edge tuple {edge} must be a 2-tuple or 3-tuple."
                )
            new_graph = new_graph.add_edge(source, sink, key, **data, **attr)
        return new_graph

    @property
    def nodes(self, data=False, default=None):
        return NodeView(self)(data=data, default=default)

    def edges(self, nbunch=None, data=False, keys=False, default=None):
        if nbunch is not None and not isinstance(nbunch, Iterable):
            if nbunch not in self:
                raise KeyError(f"{nbunch} is not a node in the graph")
        return OutMultiEdgeView(self)(
            data=data, nbunch=nbunch, keys=keys, default=default
        )

    out_edges = edges

    def in_edges(self, nbunch=None, data=False, keys=False, default=None):
        if nbunch is not None and not isinstance(nbunch, Iterable):
            if nbunch not in self:
                raise KeyError(f"{nbunch} is not a node in the graph")
        return InMultiEdgeView(self)(
            data=data, nbunch=nbunch, keys=keys, default=default
        )

    def nbunch_iter(self, nbunch=None):
        if nbunch is None:
            return iter(self)
        if not isinstance(nbunch, Iterable):
            nbunch = [nbunch]
        return (node for node in nbunch if node in self)

    def neighbors(self, node):
        return iter(self.adj[node])

    def degree(self, nbunch=None, weight=None):
        return DiMultiDegreeView(self)(nbunch, weight)

    def in_degree(self, nbunch=None, weight=None):
        return InMultiDegreeView(self)(nbunch, weight)

    def out_degree(self, nbunch=None, weight=None):
        return OutMultiDegreeView(self)(nbunch, weight)

    size = networkx.MultiDiGraph.size
    number_of_edges = networkx.MultiDiGraph.number_of_edges

    def has_successor(self, node, successor):
        return successor in self._succ[node]

    def successors(self, node):
        return iter(self._succ[node].keys())

    def has_predecessor(self, node, predecessor):
        return predecessor in self._pred[node]

    def predecessors(self, node):
        return iter(self.pred[node].keys())

    def to_undirected(self, **kwargs) -> "networkx.MultiGraph":
        return to_networkx(self).to_undirected(**kwargs)

    def reverse(self, **kwargs) -> "networkx.MultiDiGraph":
        graph = to_networkx(self)
        graph = graph.reverse(**kwargs)
        graph = from_networkx(graph)
        return graph

    def subgraph(self, nodes) -> "networkx.MultiDiGraph":
        graph = to_networkx(self)
        graph = graph.subgraph(nodes)
        graph = from_networkx(graph)
        return graph


def topological_traversal(graph):
    return networkx.topological_sort(graph)


def default_visualize_node(graphviz_graph, graph, node):
    graphviz_graph.node(node.name, label=f"{node}")


def default_visualize_edge(graphviz_graph, graph, edge):
    source, sink, keys, data = edge
    graphviz_graph.edge(
        source.name,
        sink.name,
    )


def visualize_graph(
    graph: MultiDiGraph,
    *,
    graphviz_graph=None,
    visualize_node=default_visualize_node,
    visualize_edge=default_visualize_edge,
    file_name: Optional[Union[str, pathlib.Path]] = None,
) -> None:
    if graphviz_graph is None:
        graphviz_graph = graphviz.Digraph()

    for node in graph:
        visualize_node(graphviz_graph, graph, node)

    for node in graph:
        for edge in graph.in_edges(node, data=True, keys=True):
            visualize_edge(graphviz_graph, graph, edge)

    if file_name is not None:
        file_name = pathlib.Path(file_name)
        if file_name.suffix != ".svg":
            raise ValueError(
                f"file_name must have a .svg suffix, not {file_name.suffix}"
            )
        format = file_name.suffix[1:]
        graphviz_graph.render(file_name.with_suffix(""), format=format)
        logger.info(f'Graph visualization saved to "{file_name}"')

    return graphviz_graph


def from_networkx(nx_graph: networkx.MultiDiGraph) -> MultiDiGraph:
    graph = MultiDiGraph()

    for node, data in nx_graph.nodes(data=True):
        graph = graph.add_node(node, **data)

    for source, sink, key, data in nx_graph.edges(keys=True, data=True):
        graph = graph.add_edge(source, sink, key, **data)

    graph = graph.add_attributes(**nx_graph.graph)

    return graph


def to_networkx(graph) -> "networkx.MultiDiGraph":
    nx_graph = networkx.MultiDiGraph()

    for node, data in graph.nodes(data=True):
        nx_graph.add_node(node, **data)

    for source, sink, key, data in graph.edges(keys=True, data=True):
        nx_graph.add_edge(source, sink, key, **data)

    nx_graph.graph.update(**graph.graph)
    return nx_graph


def compose_all(*graphs) -> "MultiDiGraph":
    new_graph, *_ = tuple(graphs)

    for graph in graphs:
        new_graph = new_graph.add_nodes_from(graph.nodes(data=True))

    for graph in graphs:
        new_graph = new_graph.add_edges_from(graph.edges(keys=True, data=True))

    return new_graph


def merge_graphs(*graph_node_pairs) -> "MultiDiGraph":
    def merge_binary(graph_a, graph_b: "MultiDiGraph", graph_b_node):
        graph_b_operands = list(graph_b.predecessors(graph_b_node))
        if graph_b_operands and all(operand in graph_a for operand in graph_b_operands):
            _node = graph_a._node.set(graph_b_node, graph_b._node[graph_b_node])
            new_graph = graph_a.set(_node=_node)
            for source, sink, key, data in graph_b.in_edges(
                graph_b_node, keys=True, data=True
            ):
                new_graph = new_graph.add_edge(source, sink, key, **data)
            return new_graph

        else:
            _node = graph_a._node.update(graph_b._node)

            def merge_edges(node_to_neighbors, graph_b_node_to_neighbors):
                for (
                    graph_b_node,
                    graph_b_neighbors,
                ) in graph_b_node_to_neighbors.items():
                    if graph_b_node not in node_to_neighbors:
                        node_to_neighbors = node_to_neighbors.set(
                            graph_b_node, graph_b_neighbors
                        )
                        continue

                    neighbors = node_to_neighbors[graph_b_node]
                    for graph_b_neighbor, graph_b_edges in graph_b_neighbors.items():
                        if graph_b_neighbor not in neighbors:
                            neighbors = neighbors.set(graph_b_neighbor, graph_b_edges)
                            continue

                        edges = neighbors[graph_b_neighbor].update(
                            graph_b_neighbors[graph_b_neighbor]
                        )
                        neighbors = neighbors.set(graph_b_neighbor, edges)

                    node_to_neighbors = node_to_neighbors.set(graph_b_node, neighbors)
                return node_to_neighbors

            _pred = merge_edges(graph_a._pred, graph_b._pred)
            _succ = merge_edges(graph_a._succ, graph_b._succ)

            new_graph = graph_a.set(_node=_node, _pred=_pred, _succ=_succ)
            return new_graph

    (graph, node), *graph_node_pairs = graph_node_pairs
    for other_graph, other_node in graph_node_pairs:
        graph = merge_binary(graph, other_graph, other_node)
    return graph



import torch

import torchtrail


def test_module_list():

    class Module(torch.nn.Module):

        def __init__(self):
            super(Module, self).__init__()
            self.layers = torch.nn.ModuleList(
                [torch.nn.Conv2d(3, 3, 3) for _ in range(4)]
            )

        def forward(self, x):
            for module in self.layers:
                x = module(x)
            return x

    module = Module()

    with torchtrail.trace():
        output = module(torch.rand(1, 3, 100, 100))

    torchtrail.visualize(output)
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) == 13



# MIT License

# Copyright (c) 2024 Akhmed Rakhmati

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import annotations

from contextlib import contextmanager
import dataclasses
import inspect
import math
import pathlib
import time
from typing import Any, Callable, Optional, Union, Tuple

import graphviz
from loguru import logger
import torch
from pyrsistent import PClass, field

from torchtrail import multidigraph


TORCH_NN_MODULE_CALL = torch.nn.Module.__call__


# The following functions are overriden to capture input tensors
TORCH_CREATION_OPERATION_NAMES = [
    "as_tensor",
    "from_numpy",
    "zeros",
    "zeros_like",
    "ones",
    "ones_like",
    "arange",
    "range",
    "linspace",
    "logspace",
    "eye",
    "empty",
    "empty_like",
    "full",
    "full_like",
    "complex",
    "heaviside",
    "bernoulli",
    "multinomial",
    "normal",
    "poisson",
    "rand",
    "rand_like",
    "randint",
    "randint_like",
    "randn",
    "randn_like",
    "randperm",
]
TORCH_CREATION_OPERATIONS = [
    getattr(torch, name) for name in TORCH_CREATION_OPERATION_NAMES
]


class Node(PClass):
    name = field(type=str, mandatory=True)
    unique_id = field(type=int, mandatory=True)

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        return self.unique_id < other.unique_id


class PositionalArgumentName(PClass):
    index = field(type=int, mandatory=True)

    def __repr__(self):
        return f"{self.index}"


class InputTensorIndex(PClass):
    index = field(type=int, mandatory=True)

    def __repr__(self):
        return f"${self.index}"


class TorchTensor(PClass):
    tensor: torch.Tensor = field(mandatory=True)

    def to_string(self, verbose=False):
        return "torch.Tensor"

    __repr__ = to_string


class TorchParameter(PClass):
    parameter = field(mandatory=True)

    def to_string(self, verbose=False):
        output = "torch.nn.Parameter"
        if hasattr(self.parameter, "torchtrail_name"):
            output = f"{output}\n{self.parameter.torchtrail_name}"
        return output

    __repr__ = to_string


class TorchFunction(PClass):
    function = field(mandatory=True)
    arg_name_value_pairs = field(mandatory=True)

    def to_string(self, verbose=False):
        if "__module__" in dir(self.function):
            output = f"{self.function.__module__}.{self.function.__name__}"
        elif "__objclass__" in dir(self.function):
            output = f"{self.function.__objclass__.__module__}.{ self.function.__objclass__.__name__}.{self.function.__name__}"
        elif "__class__" in dir(self.function):
            output = f"{self.function.__class__}.{self.function.__name__}"
        else:
            raise RuntimeError(f"Unknown function type: {type(self.function)}")
        output = output.replace("torch.nn.modules", "torch.nn")
        output = output.replace("torch._tensor", "torch.Tensor")
        output = output.replace("torch._C.TensorBase", "torch.Tensor")
        output = output.replace("torch._C._nn", "torch.nn.functional")
        output = output.replace("torch._C", "torch")

        if not verbose:
            return output

        output += "("
        current_length = len(output)
        if current_length > 50:
            output += "\n"
            current_length = 0
        for index, (name, value) in enumerate(self.arg_name_value_pairs):
            if isinstance(value, torch.Tensor):
                value = f"torch.Tensor(...)"
            elif isinstance(name, PositionalArgumentName):
                value_as_str = f"{value}"
            else:
                value_as_str = f"{name}={value}"
            current_length += len(value_as_str)
            output += value_as_str
            if index != len(self.arg_name_value_pairs) - 1:
                if current_length < 50:
                    output += ", "
                else:
                    output += ",\n"
                    current_length = 0

        output += ")"
        return output

    __repr__ = to_string


class TorchModule(PClass):
    module = field(mandatory=True)
    graph: multidigraph.MultiDiGraph = field(mandatory=True)
    inputs: list[TracedTensor] = field(mandatory=True)
    outputs: list[TracedTensor] = field(mandatory=True)
    arg_name_value_pairs = field(mandatory=True)

    def to_string(self, verbose=False):
        output = f"{type(self.module).__module__}.{type(self.module).__name__}"

        if verbose:
            output += "("
            current_length = len(output)
            if current_length > 50:
                output += "\n"
                current_length = 0
            for index, (name, value) in enumerate(self.arg_name_value_pairs):
                if isinstance(name, PositionalArgumentName):
                    value_as_str = f"{value}"
                else:
                    value_as_str = f"{name}={value}"

                current_length += len(value_as_str)
                output += value_as_str
                if index != len(self.arg_name_value_pairs) - 1:
                    if current_length < 50:
                        output += ", "
                    else:
                        output += ",\n"
                        current_length = 0

            output += ")"

        if self.module.torchtrail_name != "":
            output += f"\n{self.module.torchtrail_name}"
        return output

    __repr__ = to_string


class TorchModuleInput(PClass):
    name: str = field(mandatory=True)

    def to_string(self, verbose=False):
        return f"{self.name}"

    __repr__ = to_string


class ModuleIOTensor:
    def __init__(self, graph: multidigraph.MultiDiGraph, node: Node, output_index: int):
        self.graph: multidigraph.MultiDiGraph = graph
        self.node: Node = node
        self.output_index: int = output_index

    @classmethod
    def from_traced_tensor(cls, traced_tensor: TracedTorchTensor):
        return cls(traced_tensor.graph, traced_tensor.node, traced_tensor.output_index)

    @property
    def name(self) -> str:
        return self.node.name

    @property
    def shape(self) -> str:
        return self.graph.nodes[self.node]["shapes"][self.output_index]


UNIQUE_ID = 0


def get_unique_id():
    global UNIQUE_ID
    output = UNIQUE_ID
    UNIQUE_ID += 1
    return output


def create_input_tensor(
    tensor: torch.Tensor,
    function: Optional[Callable[..., Any]] = None,
    arg_name_value_pairs=None,
    duration=None,
) -> TracedTorchTensor:
    if isinstance(tensor, torch.nn.Parameter):
        unique_id = get_unique_id()
        node_name = f"torch_parameter_{unique_id}"
        node = Node(name=node_name, unique_id=unique_id)
        graph = multidigraph.MultiDiGraph().add_node(
            node,
            operation=TorchParameter(parameter=tensor),
            shapes=(tuple(tensor.shape),),
            dtypes=(tensor.dtype,),
            duration=duration,
        )
        return TracedTorchTensor(tensor, graph=graph, node=node, output_index=0)
    elif isinstance(tensor, torch.Tensor):
        if function is None:
            operation = TorchTensor(tensor=tensor)
        else:
            arg_name_value_pairs = (
                arg_name_value_pairs if arg_name_value_pairs is not None else {}
            )
            operation = TorchFunction(
                function=function, arg_name_value_pairs=arg_name_value_pairs
            )

        unique_id = get_unique_id()
        node_name = f"torch_input_{unique_id}"
        node = Node(name=node_name, unique_id=unique_id)
        graph = multidigraph.MultiDiGraph().add_node(
            node,
            operation=operation,
            shapes=(tuple(tensor.shape),),
            dtypes=(tensor.dtype,),
            duration=duration,
        )
        return TracedTorchTensor(tensor, graph=graph, node=node, output_index=0)
    else:
        raise RuntimeError(f"Unknown input type: {type(tensor)}")


def preprocess_args_and_kwargs(*function_args, **function_kwargs) -> Any:
    def preprocess_arg(arg: Any) -> Any:
        if isinstance(arg, TracedTensor):
            return arg
        elif isinstance(arg, torch.Tensor):
            return create_input_tensor(arg)
        elif isinstance(arg, (tuple, list)):
            return type(arg)([preprocess_arg(element) for element in arg])
        elif isinstance(arg, dict):
            return {key: preprocess_arg(value) for key, value in arg.items()}
        else:
            return arg

    function_args = [preprocess_arg(arg) for arg in function_args]
    function_kwargs = {
        name: preprocess_arg(arg) for name, arg in function_kwargs.items()
    }
    return function_args, function_kwargs


def get_input_tensors(object):
    input_tensors = []
    if isinstance(object, TracedTensor):
        input_tensors.append(object)
    elif isinstance(object, (list, tuple)):
        for element in object:
            input_tensors += get_input_tensors(element)
    elif isinstance(object, dict):
        for value in object.values():
            input_tensors += get_input_tensors(value)
    return input_tensors


def get_arg_names(function, *, function_args, function_kwargs):
    if inspect.isbuiltin(function):
        arg_names = [
            PositionalArgumentName(index=index) for index in range(len(function_args))
        ] + list(function_kwargs.keys())
    elif inspect.ismethoddescriptor(function):
        arg_names = [
            PositionalArgumentName(index=index) for index in range(len(function_args))
        ] + list(function_kwargs.keys())
    else:
        signature = inspect.signature(function)
        arg_names = [parameter.name for parameter in signature.parameters.values()]
        if inspect.ismethod(function):
            arg_names = ["self"] + arg_names
        if len(arg_names) < len(function_args) + len(function_kwargs):
            arg_names = [
                PositionalArgumentName(index=index)
                for index in range(len(function_args))
            ] + list(function_kwargs.keys())

    if len(arg_names) < len(function_args) + len(function_kwargs):
        raise RuntimeError(
            f"Number of argument names must be at least as large as the number of arguments"
        )
    return arg_names


def get_arg_name_value_pairs(function, *, function_args, function_kwargs):
    arg_names = get_arg_names(
        function, function_args=function_args, function_kwargs=function_kwargs
    )

    try:
        signature = inspect.signature(function)
    except:
        signature = None

    input_tensor_index = 0

    def process_arg_value(arg_value):
        nonlocal input_tensor_index
        if isinstance(arg_value, TracedTensor):
            output = InputTensorIndex(index=input_tensor_index)
            input_tensor_index += 1
            return output
        elif isinstance(arg_value, (tuple, list)):
            return type(arg_value)(
                [process_arg_value(element) for element in arg_value]
            )
        elif isinstance(arg_value, dict):
            return {key: process_arg_value(value) for key, value in arg_value.items()}
        else:
            return arg_value

    arg_name_value_pairs = []
    for arg_name, arg_value in zip(arg_names, function_args):

        if signature is not None:
            if (
                arg_name in signature.parameters
                and signature.parameters[arg_name].default != inspect.Parameter.empty
                and arg_value == signature.parameters[arg_name].default
            ):
                continue
        arg_name_value_pairs.append((arg_name, process_arg_value(arg_value)))

    for arg_name in arg_names[len(function_args) :]:
        if arg_name not in function_kwargs:
            continue
        arg_value = function_kwargs[arg_name]

        if signature is not None:
            if (
                arg_name in signature.parameters
                and signature.parameters[arg_name].default != inspect.Parameter.empty
                and arg_value == signature.parameters[arg_name].default
            ):
                continue
        arg_name_value_pairs.append((arg_name, process_arg_value(arg_value)))

    return arg_name_value_pairs


def preprocess_return_value(return_value):
    output_tensors = []
    if isinstance(
        return_value,
        (
            int,
            torch.Size,
            torch.device,
            torch.dtype,
            str,
        ),
    ):
        pass
    elif isinstance(return_value, TracedTensor):
        output_tensors.append(return_value)
    elif isinstance(return_value, torch.Tensor):
        output_tensors.append(create_input_tensor(return_value))
    elif isinstance(return_value, (tuple, list)):
        for value in return_value:
            output_tensors += preprocess_return_value(value)
    elif dataclasses.is_dataclass(return_value):
        for class_field in dataclasses.fields(return_value):
            value = getattr(return_value, class_field.name)
            output_tensors += preprocess_return_value(value)
    elif isinstance(return_value, dict):
        for value in return_value.values():
            output_tensors += preprocess_return_value(value)
    elif return_value is None:
        pass
    else:
        logger.warning(
            f"preprocess_return_value: unsupported type {type(return_value)}"
        )
    return output_tensors


def postprocess_return_value(return_value, output_tensors):
    if isinstance(return_value, TracedTensor):
        output_tensor, *_ = output_tensors
        output_tensors.pop(0)
        return output_tensor
    elif isinstance(return_value, torch.Tensor):
        output_tensor, *_ = output_tensors
        output_tensors.pop(0)
        return output_tensor
    elif isinstance(return_value, (tuple, list)):
        return type(return_value)(
            postprocess_return_value(value, output_tensors) for value in return_value
        )
    elif dataclasses.is_dataclass(return_value):
        updated_fields = {}
        for class_field in dataclasses.fields(return_value):
            value = getattr(return_value, class_field.name)
            updated_fields[class_field.name] = postprocess_return_value(
                value, output_tensors
            )
        return type(return_value)(**updated_fields)
    elif isinstance(return_value, dict):
        return {
            name: postprocess_return_value(value, output_tensors)
            for name, value in return_value.items()
        }
    else:
        return return_value


class TracedTensor: ...


class TracedTorchTensor(torch.Tensor, TracedTensor):
    @staticmethod
    def __new__(
        cls: Any,
        tensor: Any,
        graph: multidigraph.MultiDiGraph,
        node: Node,
        output_index: int,
        *function_args: Any,
        **function_kwargs: Any,
    ) -> Any:
        return super().__new__(cls, tensor, *function_args, **function_kwargs)  # type: ignore[call-arg]

    def __init__(
        self,
        tensor: Any,
        *,
        graph: multidigraph.MultiDiGraph,
        node: Node,
        output_index: int,
    ):
        self.graph: multidigraph.MultiDiGraph = graph
        self.node: Node = node
        self.output_index: int = output_index

    @property
    def name(self) -> str:
        return self.node.name

    @classmethod
    def __torch_function__(
        cls: Any,
        function,
        types: Any,
        function_args: Any = (),
        function_kwargs: Any = None,
    ) -> Any:

        if function_kwargs is None:
            function_kwargs = {}

        function_args, function_kwargs = preprocess_args_and_kwargs(
            *function_args, **function_kwargs
        )
        input_tensors = get_input_tensors(function_args) + get_input_tensors(
            function_kwargs
        )

        start_time = time.time()
        function_return_value = super().__torch_function__(
            function, types, function_args, function_kwargs
        )
        end_time = time.time()
        duration = end_time - start_time

        output_tensors = preprocess_return_value(function_return_value)
        if not output_tensors:
            return function_return_value

        shapes = tuple(tuple(tensor.shape) for tensor in output_tensors)
        dtypes = tuple(tensor.dtype for tensor in output_tensors)

        arg_name_value_pairs = get_arg_name_value_pairs(
            function, function_args=function_args, function_kwargs=function_kwargs
        )

        unique_id = get_unique_id()
        node_name = f"{function.__name__}_{unique_id}"
        node = Node(name=node_name, unique_id=unique_id)
        graph = multidigraph.merge_graphs(
            *((tensor.graph, tensor.node) for tensor in input_tensors)
        )
        graph = graph.add_node(
            node,
            operation=TorchFunction(
                function=function, arg_name_value_pairs=arg_name_value_pairs
            ),
            shapes=shapes,
            dtypes=dtypes,
            duration=duration,
        )
        for input_index, tensor in enumerate(input_tensors):
            graph = graph.add_edge(
                tensor.node,
                node,
                source_output_index=tensor.output_index,
                sink_input_index=input_index,
            )

        output_tensors = [
            TracedTorchTensor(tensor, graph=graph, node=node, output_index=output_index)
            for output_index, tensor in enumerate(output_tensors)
        ]
        return postprocess_return_value(function_return_value, output_tensors)


def wrap_create_function(function: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(*function_args: Any, **function_kwargs: Any) -> TracedTensor:
        arg_name_value_pairs = get_arg_name_value_pairs(
            function, function_args=function_args, function_kwargs=function_kwargs
        )
        start_time = time.time()
        input_tensor = function(*function_args, **function_kwargs)
        end_time = time.time()
        duration = end_time - start_time
        return create_input_tensor(
            input_tensor,
            function,
            arg_name_value_pairs=arg_name_value_pairs,
            duration=duration,
        )

    return wrapper


def create_module_input(name, tensor: torch.Tensor) -> TracedTensor:
    unique_id = get_unique_id()
    node_name = f"module_input_{unique_id}"
    node = Node(name=node_name, unique_id=unique_id)
    graph = multidigraph.MultiDiGraph().add_node(
        node,
        operation=TorchModuleInput(name=name),
        shapes=(tuple(tensor.shape),),
        dtypes=(tensor.dtype,),
    )
    return TracedTorchTensor(tensor, graph=graph, node=node, output_index=0)


def convert_to_module_args_and_kwargs(module, *function_args, **function_kwargs) -> Any:
    def preprocess_arg(name: str, arg: Any) -> Any:
        if isinstance(arg, TracedTensor):
            output = create_module_input(name, arg)
            return output
        elif isinstance(arg, torch.nn.Parameter):
            raise RuntimeError("Module parameters are not supported")
        elif isinstance(arg, (tuple, list)):
            return type(arg)([preprocess_arg(name, element) for element in arg])
        elif isinstance(arg, dict):
            return {key: preprocess_arg(name, element) for key, element in arg.items()}
        else:
            return arg

    arg_names = get_arg_names(
        module.forward, function_args=function_args, function_kwargs=function_kwargs
    )
    function_args = [
        preprocess_arg(name, arg) for name, arg in zip(arg_names, function_args)
    ]
    function_kwargs = {
        name: preprocess_arg(name, arg) for name, arg in function_kwargs.items()
    }
    return function_args, function_kwargs


def create_module(
    module, module_input_tensors, module_output_tensors, arg_name_value_pairs
):
    module_inputs = [
        ModuleIOTensor.from_traced_tensor(tensor) for tensor in module_input_tensors
    ]
    module_outputs = [
        ModuleIOTensor.from_traced_tensor(tensor) for tensor in module_output_tensors
    ]
    module_graph = multidigraph.compose_all(
        *[tensor.graph for tensor in module_inputs + module_outputs]
    )
    operation = TorchModule(
        module=module,
        graph=module_graph,
        inputs=module_inputs,
        outputs=module_outputs,
        arg_name_value_pairs=arg_name_value_pairs,
    )
    return operation


def set_torchtrail_name(module, name):
    if hasattr(module, "torchtrail_name"):
        return
    module.torchtrail_name = name

    if isinstance(module, torch.nn.ModuleList):
        for index, child in enumerate(module.children()):
            set_torchtrail_name(child, f"{name}.{index}" if name else f"{index}")

    for child_name, child in module.named_children():
        set_torchtrail_name(child, f"{name}.{child_name}" if name else child_name)

    for parameter_name, parameter in module.named_parameters():
        parameter.torchtrail_name = parameter_name


def traced_module_forward(*function_args: Any, **function_kwargs: Any) -> Any:
    module = function_args[0]
    set_torchtrail_name(module, "")

    function_args, function_kwargs = preprocess_args_and_kwargs(
        *function_args, **function_kwargs
    )

    module_args, module_kwargs = convert_to_module_args_and_kwargs(
        module, *function_args, **function_kwargs
    )

    start_time = time.time()
    module_return_value = TORCH_NN_MODULE_CALL(*module_args, **module_kwargs)
    end_time = time.time()
    duration = end_time - start_time

    module_input_tensors = get_input_tensors(module_args) + get_input_tensors(
        module_kwargs
    )
    module_output_tensors = preprocess_return_value(module_return_value)
    input_tensors = get_input_tensors(function_args) + get_input_tensors(
        function_kwargs
    )
    graph = multidigraph.merge_graphs(
        *((tensor.graph, tensor.node) for tensor in input_tensors)
    )

    shapes = tuple(tuple(tensor.shape) for tensor in module_output_tensors)
    dtypes = tuple(tensor.dtype for tensor in module_output_tensors)

    arg_name_value_pairs = get_arg_name_value_pairs(
        module.forward, function_args=function_args, function_kwargs=function_kwargs
    )[1:]
    operation = create_module(
        module, module_input_tensors, module_output_tensors, arg_name_value_pairs
    )

    unique_id = get_unique_id()
    node_name = f"{module.torchtrail_name}_{unique_id}"
    node = Node(name=node_name, unique_id=unique_id)

    graph = graph.add_node(
        node,
        operation=operation,
        shapes=shapes,
        dtypes=dtypes,
        duration=duration,
    )

    for input_index, tensor in enumerate(input_tensors):
        graph = graph.add_edge(
            tensor.node,
            node,
            source_output_index=tensor.output_index,
            sink_input_index=input_index,
        )

    output_tensors = [
        TracedTorchTensor(tensor, graph=graph, node=node, output_index=output_index)
        for output_index, tensor in enumerate(module_output_tensors)
    ]
    return postprocess_return_value(module_return_value, output_tensors)


@contextmanager
def trace():
    try:

        # Monkey-patch module __call__ and torch creation ops
        setattr(torch.nn.Module, "__call__", traced_module_forward)

        for name, op in zip(TORCH_CREATION_OPERATION_NAMES, TORCH_CREATION_OPERATIONS):
            setattr(torch, name, wrap_create_function(op))

        yield

    finally:
        # Reset monkey-patched module __call__ and torch creation ops
        setattr(torch.nn.Module, "__call__", TORCH_NN_MODULE_CALL)

        for name, op in zip(TORCH_CREATION_OPERATION_NAMES, TORCH_CREATION_OPERATIONS):
            setattr(torch, name, op)


LEVEL_COLORS = [
    "#0000ff80",
    "#ee00ee80   ",
    "#ff000080",
    "#eeee0080",
    "#00ff0080",
    "#00eeee80",
]


def get_source(graph, node, source_output_index, level, max_depth=None):
    operation = graph.nodes[node]["operation"]

    if max_depth is not None and level + 1 == max_depth:
        return node, source_output_index, graph.nodes[node]

    if not isinstance(operation, TorchModule):
        return node, source_output_index, graph.nodes[node]

    module = operation
    module_graph = module.graph
    module_node = module.outputs[source_output_index].node
    module_node_output_index = module.outputs[source_output_index].output_index
    return get_source(
        module_graph,
        module_node,
        module_node_output_index,
        level=level + 1,
        max_depth=max_depth,
    )


def get_sink(graph, node, sink_input_index, level, max_depth=None):
    operation = graph.nodes[node]["operation"]

    if not isinstance(operation, TorchModule):
        return node, sink_input_index, graph.nodes[node]

    if max_depth is not None and level + 1 == max_depth:
        return node, sink_input_index, graph.nodes[node]

    module = operation
    module_graph = module.graph
    module_node = module.inputs[sink_input_index].node
    return get_sink(
        module_graph,
        module_node,
        0,  # module_node is always a torchtrail.tracer.TorchModuleInput
        level=level + 1,
        max_depth=max_depth,
    )


def duration_to_string(duration):
    if duration < 1e-6:
        return f"{duration * 1e9:.1f} ns"
    elif duration < 1e-3:
        return f"{duration * 1e6:.1f} µs"
    elif duration < 1:
        return f"{duration * 1e3:.1f} ms"
    else:
        return f"{duration:.1f} s"


def visualize_node(
    graphviz_graph,
    graph,
    node,
    max_depth,
    visualize_node,
    visualize_edge,
    level,
    verbose,
):
    attributes = graph.nodes[node]
    operation = attributes["operation"]

    input_tensors = []
    for source_node, _, edge_data in graph.in_edges(node, data=True):
        input_shape = graph.nodes[source_node]["shapes"][
            edge_data["source_output_index"]
        ]
        input_dtype = graph.nodes[source_node]["dtypes"][
            edge_data["source_output_index"]
        ]
        input_tensors.append((input_shape, input_dtype))

    output_shapes = attributes["shapes"]
    output_dtypes = attributes["dtypes"]
    output_tensors = tuple(
        (shape, dtype) for shape, dtype in zip(output_shapes, output_dtypes)
    )

    label = operation.to_string(verbose=verbose)

    duration = attributes.get("duration", None)
    if duration is not None:
        label = f"{label}\nduration: {duration_to_string(duration)}"

    num_columns = max(len(input_tensors), len(output_tensors))

    table = label.replace("\n", "<BR/>")
    table = f"""<
            <TABLE BORDER="{0}" CELLBORDER="{1}"
            CELLSPACING="{1}" CELLPADDING="{1}">
            <TR>
                <TD ROWSPAN="{2 if input_tensors else 1}">{table}</TD>
            """

    def compute_even_column_sizes(num_columns, num_tensors):
        if num_tensors == 0:
            return []
        column_size = math.ceil(num_columns // num_tensors)
        column_sizes = []
        remaining = num_columns
        for _ in range(num_tensors):
            if remaining > column_size:
                column_sizes.append(column_size)
            else:
                column_sizes.append(remaining)
            remaining -= column_size
        return column_sizes

    input_column_sizes = compute_even_column_sizes(num_columns, len(input_tensors))
    for index, (shape, dtype) in enumerate(input_tensors):
        column_size = input_column_sizes[index]
        table = (
            table
            + f"""
                <TD PORT="${index}" COLSPAN="{column_size}">Input {index}<BR/>{shape}<BR/>{dtype}</TD>
            """
        )
    table += """
            </TR>
            <TR>
            """
    output_column_sizes = compute_even_column_sizes(num_columns, len(output_tensors))
    for index, (shape, dtype) in enumerate(output_tensors):
        column_size = output_column_sizes[index]
        table += f"""
                <TD PORT="#{index}" COLSPAN="{column_size}">Output {index}<BR/>{shape}<BR/>{dtype}</TD>
            """
    table += f"""
            </TR>
            </TABLE>>"""

    if isinstance(operation, TorchModule):
        color = LEVEL_COLORS[level % len(LEVEL_COLORS)]
        if max_depth is None or level < max_depth - 1:
            with graphviz_graph.subgraph(name=node.name) as cluster_graph:
                cluster_graph.attr(
                    fontcolor="black",
                    bgcolor=color,
                    cluster="true",
                    label=label,
                    rankdir="TB",
                    shape="hexagon",
                )
                cluster_graph.node_attr["style"] = "filled"
                _visualize(
                    cluster_graph,
                    operation.graph,
                    max_depth=max_depth,
                    file_name=None,
                    visualize_node=visualize_node,
                    visualize_edge=visualize_edge,
                    verbose=verbose,
                    level=level + 1,
                )
        else:
            graphviz_graph.node(
                node.name,
                label=table,
                fontcolor="black",
                fillcolor=color,
            )

    else:
        graphviz_graph.node(
            node.name,
            label=table,
            fillcolor="#DCDCDC",
        )


def visualize_edge(graphviz_graph, graph, edge, max_depth, level):
    source_node, sink_node, _, edge_data = edge

    source_output_index = edge_data["source_output_index"]
    sink_input_index = edge_data["sink_input_index"]

    source_node, source_output_index, *_ = get_source(
        graph,
        source_node,
        source_output_index,
        level,
        max_depth=max_depth,
    )

    sink_node, sink_input_index, sink_attributes = get_sink(
        graph,
        sink_node,
        sink_input_index,
        level,
        max_depth=max_depth,
    )

    source_name = f"{source_node.name}:#{source_output_index}"
    sink_name = f"{sink_node.name}:${sink_input_index}"
    if isinstance(sink_attributes["operation"], TorchModuleInput):
        sink_name = sink_node.name

    graphviz_graph.edge(
        source_name,
        sink_name,
        label=f"{source_output_index} -> {sink_input_index}",
        fontcolor="black" if level == 0 else "white",
    )


def _visualize(
    graphviz_graph,
    graph,
    *,
    max_depth,
    file_name,
    visualize_node,
    visualize_edge,
    verbose,
    level=0,
) -> graphviz.Digraph:

    if max_depth is not None:
        if max_depth < 1:
            raise RuntimeError("max_depth must be greater than 0")
        if level == max_depth:
            return graphviz_graph

    for node in sorted(graph):
        visualize_node(
            graphviz_graph,
            graph,
            node,
            max_depth,
            visualize_node,
            visualize_edge,
            level,
            verbose=verbose,
        )

    for node in sorted(graph):
        for edge in graph.in_edges(node, data=True, keys=True):
            visualize_edge(graphviz_graph, graph, edge, max_depth, level)

    if file_name is not None:
        file_name = pathlib.Path(file_name)
        if file_name.suffix not in {".svg", ".png", ".pdf"}:
            raise ValueError(
                f"file_name must have a .svg, .png or .pdf suffix, not {file_name.suffix}"
            )
        format = file_name.suffix[1:]
        graphviz_graph.render(file_name.with_suffix(""), format=format)
        logger.info(f'Graph visualization saved to "{file_name}"')

    return graphviz_graph


def _flatten_graph(
    graph,
    *,
    new_graph=None,
    level=0,
) -> multidigraph.MultiDiGraph:

    if new_graph is None:
        new_graph = multidigraph.MultiDiGraph()

    for node in multidigraph.topological_traversal(graph):
        operation = graph.nodes[node]["operation"]
        if isinstance(operation, TorchModule):
            module_graph = _flatten_graph(
                operation.graph, new_graph=new_graph, level=level + 1
            )
            new_graph = multidigraph.compose_all(new_graph, module_graph)
        else:
            new_graph = new_graph.add_node(
                node,
                **graph.nodes[node],
            )

    for node in multidigraph.topological_traversal(graph):
        operation = graph.nodes[node]["operation"]
        for source_node, sink_node, edge_data in graph.in_edges(node, data=True):
            source_node, source_output_index, _ = get_source(
                graph, source_node, edge_data["source_output_index"], level
            )
            sink_node, sink_input_index, _ = get_sink(
                graph, sink_node, edge_data["sink_input_index"], level
            )
            new_graph = new_graph.add_edge(
                source_node,
                sink_node,
                source_output_index=source_output_index,
                sink_input_index=sink_input_index,
            )

    return new_graph


def _remove_module_forward_args(graph):
    new_graph = multidigraph.MultiDiGraph()

    for node in graph:
        operation = graph.nodes[node]["operation"]
        if isinstance(operation, TorchModuleInput):
            continue
        new_graph = new_graph.add_node(
            node,
            **graph.nodes[node],
        )

    for node in multidigraph.topological_traversal(graph):
        operation = graph.nodes[node]["operation"]
        if isinstance(operation, TorchModuleInput):
            continue
        for source_node, _, edge_data in graph.in_edges(node, data=True):
            while isinstance(graph.nodes[source_node]["operation"], TorchModuleInput):
                ((source_node, _, source_node_edge_data),) = graph.in_edges(
                    source_node, data=True
                )
                edge_data = edge_data.set(
                    "source_output_index", source_node_edge_data["source_output_index"]
                )
            new_graph = new_graph.add_edge(
                source_node,
                node,
                **edge_data,
            )

    return new_graph


def flatten_graph(graph) -> multidigraph.MultiDiGraph:
    graph = _flatten_graph(graph)
    graph = _remove_module_forward_args(graph)
    return graph


def process_output(output):
    output_tensors = []
    if isinstance(output, TracedTensor):
        output_tensors.append(output)
    elif isinstance(output, (tuple, list)):
        for value in output:
            output_tensors += process_output(value)
    elif dataclasses.is_dataclass(output):
        for class_field in dataclasses.fields(output):
            value = getattr(output, class_field.name)
            output_tensors += process_output(value)
    elif isinstance(output, dict):
        for value in output.values():
            output_tensors += process_output(value)
    elif output is None:
        pass
    else:
        raise RuntimeError(f"Unexpected type {type(output)}")
    return output_tensors


def get_graph(
    value: Union[multidigraph.MultiDiGraph, TracedTensor, Tuple[TracedTensor, ...]],
    as_networkx: bool = False,
    flatten: bool = False,
) -> multidigraph.MultiDiGraph:
    if isinstance(value, multidigraph.MultiDiGraph):
        return value
    output_tensors = process_output(value)
    graph = multidigraph.compose_all(*[tensor.graph for tensor in output_tensors])
    if flatten:
        graph = flatten_graph(graph)
    if as_networkx:
        graph = multidigraph.to_networkx(graph)
    return graph


def visualize(
    output: Union[TracedTensor, Tuple[TracedTensor, ...]],
    *,
    show_modules: bool = True,
    max_depth: Optional[int] = None,
    file_name: Optional[Union[str, pathlib.Path]] = None,
    graph_attr: Optional[dict] = None,
    node_attr: Optional[dict] = None,
    edge_attr: Optional[dict] = None,
    visualize_node: Callable = visualize_node,
    visualize_edge: Callable = visualize_edge,
    verbose: bool = False,
) -> graphviz.Digraph:

    if not show_modules and max_depth is not None:
        raise RuntimeError("max_depth is not supported with show_modules=True")

    graph = get_graph(output)
    if not show_modules:
        graph = flatten_graph(graph)

    if graph_attr is None:
        graph_attr = {"ordering": "in", "rankdir": "TB"}

    if node_attr is None:
        node_attr = {
            "style": "filled",
            "border": "1",
            "fontsize": "10",
            "ranksep": "0.1",
            "height": "0.2",
            "fontname": "Linux libertine",
            "margin": "0",
            "shape": "box",
        }
    if edge_attr is None:
        edge_attr = {
            "fontsize": "10",
        }

    graphviz_graph = graphviz.Digraph(
        engine="dot",
        graph_attr=graph_attr,
        node_attr=node_attr,
        edge_attr=edge_attr,
    )

    return _visualize(
        graphviz_graph,
        graph,
        max_depth=max_depth,
        file_name=file_name,
        visualize_node=visualize_node,
        visualize_edge=visualize_edge,
        verbose=verbose,
    )



# MIT License

# Copyright (c) 2024 Akhmed Rakhmati

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from setuptools import setup
from setuptools import find_packages

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="torchtrail",
    version="0.0.15",
    author="Akhmed Rakhmati",
    author_email="akhmed.rakhmati@gmail.com",
    description="A library for tracing the execution of Pytorch operations and modules",
    packages=find_packages(),
    install_requires=[
        "torch>=1.3.0",
        "networkx>=3.1",
        "pyrsistent>=0.20.0",
        "graphviz>=0.20.1",
        "loguru>=0.6.0",
    ],
    extras_require={
        "dev": [
            "accelerate==0.27.0",
            "black[jupyter]==24.1.1",
            "pillow==10.2.0",
            "pytest==8.0.0",
            "torchvision==0.17.0",
            "transformers==4.37.2",
        ]
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/arakhmati/torchtrail",
    keywords=[
        "pytorch",
        "tracing",
        "tracing library",
        "tracing tool",
        "tracing module",
        "tracing operations",
        "visualize",
        "visualize pytorch",
        "visualize pytorch operations",
        "visualize pytorch modules",
    ],
)



import pytest

import torch
import transformers

import torchtrail


@pytest.mark.parametrize("show_modules", [True, False])
def test_falcon_decoder_layer(tmp_path, show_modules):

    model_name = "tiiuae/falcon-7b-instruct"
    config = transformers.FalconConfig.from_pretrained(model_name)
    model = transformers.models.falcon.modeling_falcon.FalconDecoderLayer(config)

    with torchtrail.trace():
        input_tensor = torch.rand((1, 64, 4544))
        attention_mask = torch.ones((1, 64), dtype=torch.long)
        output = model(input_tensor, alibi=None, attention_mask=attention_mask)

    torchtrail.visualize(
        output,
        show_modules=show_modules,
        file_name=tmp_path / "falcon_decoder_layer.svg",
    )
    assert len(torchtrail.get_graph(output)) == 3
    assert len(torchtrail.get_graph(output, flatten=True)) == 56


@pytest.mark.skip(reason="Test takes too long to run.")
@pytest.mark.parametrize("show_modules", [True, False])
def test_falcon(tmp_path, show_modules):

    model_name = "tiiuae/falcon-7b-instruct"
    model = (
        transformers.models.falcon.modeling_falcon.FalconForCausalLM.from_pretrained(
            model_name
        )
    )

    with torchtrail.trace():
        input_tensor = torch.randint(0, model.config.vocab_size, (1, 64))
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "falcon.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) == 1578


@pytest.mark.parametrize("show_modules", [True, False])
@pytest.mark.parametrize("num_hidden_layers", [1, 2])
@pytest.mark.parametrize("num_tokens", [1, 3])
@pytest.mark.parametrize("use_cache", [True, False])
def test_falcon7b_instruct_with_kv_cache(
    tmp_path, show_modules, num_hidden_layers, num_tokens, use_cache
):

    model_version = "tiiuae/falcon-7b-instruct"

    tokenizer = transformers.AutoTokenizer.from_pretrained(model_version)

    config = transformers.models.falcon.modeling_falcon.FalconConfig.from_pretrained(
        model_version
    )
    if num_hidden_layers is not None:
        config.num_hidden_layers = num_hidden_layers

    model = transformers.models.falcon.modeling_falcon.FalconForCausalLM(
        config=config
    ).eval()

    prompt_text = ["Display a graph of the falcon model"]

    tokenized_inputs = tokenizer(
        prompt_text, padding=False, add_special_tokens=False, return_tensors="pt"
    )
    input_tokens = tokenized_inputs["input_ids"]

    def post_process(logits):
        next_token_logits = logits[:, -1, :]
        next_tokens = torch.argmax(next_token_logits, dim=-1)
        tokens = next_tokens[:, None]
        return tokens

    kv_cache = None
    with torchtrail.trace():
        tokens = input_tokens
        for _ in range(num_tokens):
            model_output = model(tokens, past_key_values=kv_cache, use_cache=use_cache)
            tokens = post_process(logits=model_output.logits)
            kv_cache = model_output.past_key_values
            assert hasattr(tokens, "graph")

    torchtrail.visualize(
        tokens,
        file_name=tmp_path / "falcon7b_instruct_with_kv_cache.svg",
        show_modules=show_modules,
    )

    if num_hidden_layers == 1 and num_tokens == 1 and use_cache:
        assert len(torchtrail.get_graph(tokens)) == 5
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 62
    elif num_hidden_layers == 1 and num_tokens == 1 and not use_cache:
        assert len(torchtrail.get_graph(tokens)) == 5
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 62
    elif num_hidden_layers == 1 and num_tokens == 3 and use_cache:
        assert len(torchtrail.get_graph(tokens)) == 13
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 188
    elif num_hidden_layers == 1 and num_tokens == 3 and not use_cache:
        assert len(torchtrail.get_graph(tokens)) == 13
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 184
    elif num_hidden_layers == 2 and num_tokens == 1 and use_cache:
        assert len(torchtrail.get_graph(tokens)) == 5
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 111
    elif num_hidden_layers == 2 and num_tokens == 1 and not use_cache:
        assert len(torchtrail.get_graph(tokens)) == 5
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 111
    elif num_hidden_layers == 2 and num_tokens == 3 and use_cache:
        assert len(torchtrail.get_graph(tokens)) == 13
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 339
    elif num_hidden_layers == 2 and num_tokens == 3 and not use_cache:
        assert len(torchtrail.get_graph(tokens)) == 13
        assert len(torchtrail.get_graph(tokens, flatten=True)) == 331
    else:
        raise ValueError(
            f"Unexpected combination of num_hidden_layers={num_hidden_layers}, num_tokens={num_tokens}, use_cache={use_cache}"
        )



import pytest

import torch
import transformers

import torchtrail


@pytest.mark.parametrize("show_modules", [True, False])
def test_bert_attention(tmp_path, show_modules):
    model_name = "google/bert_uncased_L-4_H-256_A-4"
    config = transformers.BertConfig.from_pretrained(model_name)
    model = transformers.models.bert.modeling_bert.BertAttention(config).eval()

    with torchtrail.trace():
        input_tensor = torch.rand(1, 64, config.hidden_size)
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "bert_attention.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) == 33


@pytest.mark.parametrize("show_modules", [True, False])
def test_bert(tmp_path, show_modules):
    model_name = "google/bert_uncased_L-4_H-256_A-4"
    model = transformers.BertModel.from_pretrained(model_name).eval()

    with torchtrail.trace():
        input_tensor = torch.randint(0, model.config.vocab_size, (1, 64))
        output = model(input_tensor)

    torchtrail.visualize(
        output, show_modules=show_modules, file_name=tmp_path / "bert.svg"
    )
    assert len(torchtrail.get_graph(output)) == 2
    assert len(torchtrail.get_graph(output, flatten=True)) == 205
